package dev.peerat.mapping;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.RoundEnvironment;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.TypeElement;
import javax.tools.Diagnostic.Kind;
import javax.tools.FileObject;
import javax.tools.StandardLocation;

import dev.peerat.parser.ElementBuilder.Builder;
import dev.peerat.parser.Parser;
import dev.peerat.parser.TokenValidator;
import dev.peerat.parser.java.Annotation;
import dev.peerat.parser.java.Class;
import dev.peerat.parser.java.Function;
import dev.peerat.parser.java.JavaElement;
import dev.peerat.parser.java.JavaFile;
import dev.peerat.parser.java.JavaParser;
import dev.peerat.parser.java.operation.ReturnOperation;

public class TreasureProcessor extends AbstractProcessor{
	
	@Override
	public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv){
		
		String[] sources = processingEnv.getOptions().get("treasure.source").split(";");
		
		List<File> workingDir = new ArrayList<>();
		File output;
		
		try{
			FileObject resource = processingEnv.getFiler().getResource(StandardLocation.SOURCE_OUTPUT, "", "");
			output = new File(resource.toUri());
			
			for(String source : sources){
				File given = new File(source);
				if(given.equals(output)) continue;
				workingDir.add(given);
			}
		}catch(Exception e){
			return false;
		}
		
		for(File workingDirectory : workingDir){
			try {
				parseClazz(workingDirectory, output, workingDirectory);
			}catch(Exception ex){
				annotations.forEach((a) -> roundEnv.getElementsAnnotatedWith(a).forEach((e) -> processingEnv.getMessager().printMessage(Kind.ERROR, ex.getMessage(), e)));
			}
		}
	    return true;
	  }
	
	private static final Parser<JavaElement> PARSER = new JavaParser();
	
	private void parseClazz(File source, File output, File file) throws Exception{
		if(file.isDirectory()){
			for(File child : file.listFiles()) parseClazz(source, output, child);
			return;
		}
		if(!file.getName().endsWith(".java")) return;
		BufferedReader reader = new BufferedReader(new FileReader(file));
		try {
			JavaFile javaFile = new JavaFile();
			TokenValidator.TOKENS = 0;
			TokenValidator.MAX_VALIDATE = 0;
			PARSER.parse(reader, javaFile);
			process(source, output, file, javaFile);
		}catch(Exception ex){
			error(source, output, file,  new Exception(file.getName()+" | "+ex.getMessage()));
		}
	}
	
	private void process(File source, File output,  File clazz, JavaFile file) throws Exception{
		File out = new File(output, clazz.getAbsolutePath().substring(source.getAbsolutePath().length()));
		File parent = out.getParentFile();
		if(!parent.exists()) parent.mkdirs();
		if(!out.exists()) out.createNewFile();
		if(TokenValidator.TOKENS != TokenValidator.MAX_VALIDATE) throw new Exception(TokenValidator.TOKENS +" < "+TokenValidator.MAX_VALIDATE);

		BufferedWriter writer = new BufferedWriter(new FileWriter(out));
//		PARSER.build(writer);
		
		Class treasure = file.find((element) -> {
			if(element instanceof Class){
				Class c = (Class)element;
				return c.find((subelement) -> {
					if(subelement instanceof Annotation){
						Annotation annotation = (Annotation)subelement;
						if(annotation.getName().getValue().equals("Treasure")) return true; //can also be dev.peerat.mapping.Treasure
					}
					return false;
				}) != null;
			}
			return false;
		});
		if(treasure != null){
			processTreasure(file, treasure);
		}else{
			processOthers(file);
		}
		Builder builder = new Builder();
		file.build(builder);
		builder.build(writer);
		
	}
	
	private void processTreasure(JavaFile file, Class clazz){
		for(JavaElement child : clazz.getElements()){
			if(child instanceof Function){
				Function function = (Function)child;
				if(function.find((subelement) -> {
					if(subelement instanceof Annotation){
						Annotation annotation = (Annotation)subelement;
						if(annotation.getName().getValue().equals("SeaBottle")) return true; //can also be dev.peerat.mapping.SeaBottle
					}
					return false;
				}) != null){
					processSeaBottle(function);
				}
			}
		}
	}
	
	private void processSeaBottle(Function function){
		if(function.getElements().size() != 1) return;
		JavaElement target = function.getElements().get(0);
		if(!(target instanceof ReturnOperation)) return;
		ReturnOperation operation = (ReturnOperation)target;
		
	}
	
	private void processOthers(JavaFile file){
		
	}

	private void error(File source, File output,  File clazz, Exception e) throws Exception{
		File out = new File(output, "log.txt");
		File parent = out.getParentFile();
		if(!parent.exists()) parent.mkdirs();
		if(!out.exists()) out.createNewFile();
		BufferedWriter writer = new BufferedWriter(new FileWriter(out));
		writer.write(clazz.getAbsolutePath()+"\n");
		writer.write(e.getMessage()+"\n");
		for(StackTraceElement trace : e.getStackTrace()){
			writer.write(trace+"\n");
		}
		writer.flush();
		writer.close();
	}
	
	@Override
	public Set<String> getSupportedAnnotationTypes(){
		return new HashSet<>(Arrays.asList(Treasure.class.getCanonicalName()));
	}
	
	@Override
	public SourceVersion getSupportedSourceVersion(){
		return SourceVersion.RELEASE_8;
	}
}